package scala.tools.nsc.effects

import scala.tools.nsc._

abstract class EffectDomain extends Infer with RelEffects with DefaultEffects with AnfTransform {
  // `global` should not be a class parameter. Having it a field allows to refine
  // its type, e.g. `EffectDomain { val global: some.global.type }`, which is not
  // possible for parameters.
  val global: Global

  val lattice: EffectLattice
  import lattice._

  import global._

  /**
   * A flag indicating if trees should be translated to ANF [*] before invoking
   * `computeEffect`. This flag can be overridden by concrete effect domains which
   * require trees to be in ANF, such as the purity checker.
   *
   * [*] Administrative Normal Form, see "The Essence of Compiling with Continuations"
   */
  val requireANF: Boolean = false

  /**
   * The annotation classes which are used by this effect domain to annotate effects.
   */
  val annotationClasses: List[Symbol]

  val uncheckedFunctions: List[String] = List[String]()

  /**
   * Read the effect from a list of annotations. If no effect annotation for this domain
   * can be found, this method should return the `default` effect.
   *
   * This method does not need to handle `@pure` annotations: this is already done by
   * `fromAnnotations`.
   */
  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect

  /**
   * TODO: rename to be more in line with `parseAnnotationInfos`.
   * Maybe rename both. (`parseEffectFromAnnotations`, `emitEffectAsAnnotations`)
   */
  def toAnnotation(eff: Effect): List[AnnotationInfo]

  def toAnnotationUnknown(eff: Effect): Option[List[AnnotationInfo]] = {
    if (eff.unknown) Some(List(AnnotationInfo(unknownClass.tpe, Nil, Nil))) else None
  }

  /**
   * The effect represented by the annotations on the (return type of the potential method type) `tpe`.
   *
   * If there is no effect annotation for the specific domain, the default effect is used:
   *  - `bottom` if there is a `@pure` annotation
   *  - `top` otherwise
   */
  final def fromAnnotation(tpe: Type): Effect = fromAnnotationList(tpe.finalResultType.annotations)

  /**
   * The effect represented by `annotations`
   */
  final def fromAnnotationList(annots: List[AnnotationInfo]) = {
    val hasPure = annots.exists(_.atp.typeSymbol == pureClass)
    val hasUnknown = annots.exists(_.atp.typeSymbol == unknownClass)
    /*TODO: definir cual es el estado por defecto si no viene anotaciones (aca es top */
    val default = if(hasUnknown) unknown else{ if (hasPure) effectForPureAnnotated else top }

    parseAnnotationInfos(annots, default)
  }

  /**
   * True there exists some effect annotations in the list `annotations`.
   */
  final def existsEffectAnnotation(annotations: List[AnnotationInfo]) = {
    annotations.exists(ann => allEffectAnnots.contains(ann.atp.typeSymbol))
  }
  
  final def hasUncheckedAnnotation(annotations: List[AnnotationInfo]) = {
    annotations.exists(ann => ann.atp.typeSymbol == definitions.UncheckedClass)
  }

  final lazy val pureClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.pure"))

  final lazy val unknownClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.unknown"))

  final lazy val allEffectAnnots = pureClass :: unknownClass :: annotationClasses


  def neededEffect(tree: Tree): Option[lattice.Effect] = None


  //In case of no anotations, this should come by default
  def updateEffectDef(tree: Tree, effect: lattice.Effect, rtpe: Type): Option[lattice.Effect] = None

  def updateEffect(tree: Tree, eff: lattice.Effect, rtpe: Type): lattice.Effect = eff

  def annotateEffect(tree: Tree): Option[lattice.Effect] = None
  /**
   * The effect that is assigned to getters and setters. This method can be overridden
   * by effect domains to assign a different effect to accessors.
   */
  def accessorEffect(sym: Symbol, tpe: Type, tree: Tree): Effect = bottom


 case class Parent(tree: Tree) 

  def setParent(tree: Tree) = {
    tree.children.map{child => 
      child.updateAttachment[Parent](Parent(tree))
    }
  }
  def getParent(tree: Tree): Option[Parent] = {
    tree.attachments.get[Parent]
  }

  def within(tree: Tree, path: String, i: Int = 0): Boolean = {
    if(tree.symbol != null){
      /*println("trying "+buildPath(tree.symbol), "i:", i)
      println("has?", (path.r.findAllIn(buildPath(tree.symbol)).hasNext))
      println("parent?", getParent(tree))*/

      (path.r.findAllIn(buildPath(tree.symbol)).hasNext || 
      getParent(tree).map(p => within(p.tree, path, i+1)).getOrElse(false))
    } else {
      //println("symbol is null, but do I have a parent?", getParent(tree))
      getParent(tree).map(p => within(p.tree, path, i+1)).getOrElse(false)
    }
  }


}

trait WithUnknownEffect{
  val unknown: Boolean = false
}

trait EffectLattice {

  import annotation.effects._
  type Effect <: WithUnknownEffect

  def top: Effect
  def bottom: Effect

  def unknown: Effect


  /**
   * This effect is used by the framework for methods that have a `@pure` annotation.
   * By default, pure methods have the `bottom` effect. Effect domains can override
   * this method to assign a different default effect to `@pure` methods.
   */
  def effectForPureAnnotated: Effect = bottom

  def joinUnknown(a: Effect, b: Effect): Boolean = a.unknown || b.unknown
  def meetUnknown(a: Effect, b: Effect): Boolean = a.unknown && b.unknown
  def diffUnknown(a: Effect, b: Effect): Boolean = !b.unknown && a.unknown
  /*def joinUnknown(a: Effect, b: Effect): Option[Effect] = {
    if(a.unknown || b.unknown) Some(unknown)
    else None
  }
  def meetUnknown(a: Effect, b: Effect): Option[Effect] = joinUnknown(a,b)*/
  def lteUnknown(a: Effect, b: Effect): Boolean = {
    if(a.unknown || b.unknown) true
    else false
  }




  def join(a: Effect, b: Effect): Effect
  def meet(a: Effect, b: Effect): Effect

  def lte(a: Effect, b: Effect): Boolean

  final def joinAll(effs: Effect*): Effect = {
    if (effs.isEmpty) bottom
    else {
      var acc = bottom
      for (e <- effs) { acc = join(acc, e) }
      acc
    }
  }

  final def meetAll(effs: Effect*): Effect = {
    if (effs.isEmpty) top
    else {
      var acc = top
      for (e <- effs) { acc = meet(acc, e) }
      acc
    }
  }

  implicit class EffectOps(eff: Effect) {
    def u(other: Effect): Effect = join(eff, other)
    def n(other: Effect): Effect = meet(eff, other)
    def <=(other: Effect): Boolean = lte(eff, other)
  }
}
