package scala.tools.nsc.effects

import scala.tools.nsc.Global
import scala.tools.nsc.plugins.Plugin

class EffectsPlugin(val global: Global) extends Plugin {
  /** The name of this plugin. */
  val name = "effects"

  /** A short description of the plugin */
  val description = "tracks side effects"

  /**
   * Do nothing, this method is invoked too late.
   *
   * The compiler accesses the `components` filed of the compiler plugin before calling
   * `processOptions`. Therefore the `settings` object is not yet initialized when the
   * component (i.e. the `effectChecker`, and the `domain`) are computed.
   *
   * Instead of processing options here, the constructor of `settings` reads them directly
   * from `global`.
   */
  override def processOptions(options: List[String], error: String => Unit) {
    ()
  }
  
  def mkDomains(names: List[String]): List[EffectDomain { val global: EffectsPlugin.this.global.type }] = names match {
    case "io" :: xs =>
      new io.IODomain {
        val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
        override val uncheckedFunctions = efftpSettings.unchecks
      } :: mkDomains(xs)
      
    case "exceptions" :: xs =>
      new exceptions.ExceptionsDomain {
        val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
        override val uncheckedFunctions = efftpSettings.unchecks
      } :: mkDomains(xs)

    case "purity" :: xs =>
      new purity.PurityDomain {
        val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
        override val uncheckedFunctions = efftpSettings.unchecks
      } :: mkDomains(xs)

    case "ref" :: xs =>
      new ref.RefDomain {
        val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
        override val uncheckedFunctions = efftpSettings.unchecks
      } :: mkDomains(xs)

    case "example" :: xs =>
      new example.ExampleDomain {
        val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
        override val uncheckedFunctions = efftpSettings.unchecks
      } :: mkDomains(xs)

    case "myIO" :: xs =>
      new myio.MyIODomain {
        val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
        override val uncheckedFunctions = efftpSettings.unchecks
    } :: mkDomains(xs)

    case "DBAccess" :: xs =>
      new dbaccess.DBAccessDomain {
          val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
          override val uncheckedFunctions = efftpSettings.unchecks
        } :: mkDomains(xs)
    case "simpleIO" :: xs =>
      new simpleio.SimpleIODomain {
          val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
          override val uncheckedFunctions = efftpSettings.unchecks
        } :: mkDomains(xs)
    case "simpleSEIO" :: xs =>
      new simpleseio.SimpleSEIODomain {
          val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
          override val uncheckedFunctions = efftpSettings.unchecks
        } :: mkDomains(xs)
    case "simpleTPIO" :: xs =>
      new simpletpio.SimpleTPIODomain {
          val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
          override val uncheckedFunctions = efftpSettings.unchecks
        } :: mkDomains(xs)
    case x :: xs =>
      global.abort(s"Unknown effect domain: $x")
      
    case Nil => Nil

  }
  
  val domain: EffectDomain { val global: EffectsPlugin.this.global.type } = {
    mkDomains(efftpSettings.domains) match {
      case Nil =>
        global.abort(s"No effect domain specified")
        
      case x :: xs =>
        (x /: xs) {
          case (domain1, domain2) => new {
            val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
            val d1 = domain1
            val d2 = domain2
          } with BiEffectDomain
        }
    }
  }

  /* Create the TypeCheckerPlugin instance. Its constructor will register the analyzer plugin
   * and the annotation checker. No compiler phase is needed.
   */
  val typerPlugin = new {
    val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
    val domain = EffectsPlugin.this.domain
  } with TypeCheckerPlugin

//  disabled for now - maybe useful if we end up having issues in later phases
//  val cleanupPhase = new {
//    val global: EffectsPlugin.this.global.type = EffectsPlugin.this.global
//    val typeCheckerPlugin = typerPlugin
//  } with EffectsCleanup

  /**
   * The compiler components that will be applied when running this plugin
   */
  val components = Nil // List(cleanupPhase)



  object efftpSettings {
    var domains: List[String] = Nil
    var unchecks: List[String] = Nil
    val domainsStr = "domains:"
    val unchecksStr = "unchecks:"

    process(globalOptions, s => global.abort(s))

    /**
     * Read the options for this compiler plugin - see documentation on `processOptions`
     * above.
     */
    def globalOptions = {
      val namec = name+":"
      val opts = global.settings.pluginOptions.value.filter(_ startsWith namec)
      opts.map(_ stripPrefix namec)
    }

    def process(opts: List[String], error: String => Unit) {
      opts match {
        case x :: xs if x.startsWith(domainsStr) =>
          if (domains.nonEmpty) {
            error("Multiple 'domain' settings")
          } else {
            domains = x.drop(domainsStr.length).split(":").toList
          }
          process(xs, error)
        case x :: xs if x.startsWith(unchecksStr) =>
          if (unchecks.nonEmpty) {
            error("Multiple 'uncheck' settings")
          } else {
            unchecks = x.drop(unchecksStr.length).split(":").toList.map{_.replace(".","\\.").replace("*", ".*")}
          }
          process(xs, error)
        case x :: xs =>
          error(s"Unknown setting: $x")

        case Nil =>
          ()
      }
    }
  }
}

/* Some globally useful things */
object EffectsPlugin {
  lazy val traceAnf = sys.props.isDefinedAt("efftp.traceAnf")

  // debugging stuff
  def printRes[T](res: T, msg: String = ""): T = {
    print(msg)
    println(res)
    res
  }
}
