package scala.tools.nsc.effects



trait Infer { self: EffectDomain =>
  import global._
  import lattice._

  /**
   * This class is used to report effect mismatch errors.
   */
  abstract class EffectReporter {
    def issueError(tree: Tree, msg: String): Unit
    def setError(tree: Tree): Unit

    private def reportError(tree: Tree, msg: String) {
      issueError(tree, msg)
      setError(tree)
    }

    private def mismatchMsg(expected: Effect, found: Effect, detailsMsg: Option[String]) =
      "effect type mismatch;\n found   : " + found + "\n required: " + expected +
      explainMismatch(expected, found).map("\n"+_).getOrElse("") +
      detailsMsg.map("\n"+_).getOrElse("")

    def error(expected: Effect, found: Effect, tree: Tree, detailsMsg: Option[String]) {
      reportError(tree, mismatchMsg(expected, found, detailsMsg))
    }
  }
  
  def explainMismatch(expected: Effect, found: Effect): Option[String] = None

  /**
   * The context which is passed through effect inference.
   *
   * @param expected  The expected effect. If defined, an error message is issued if the inferred effect does not
   *                  conform to the expected effect.
   * @param relEnv    The relative effects which are active for inferring the effect of the expression.
   * @param reporter  The error reporter.
   * @param errorInfo An optional message which is //printed together with every effect mismatch error message.
   */
  case class EffectContext(expected: Option[Effect],
                           relEnv: List[RelEffect],
                           reporter: EffectReporter,
                           errorInfo: Option[String],
                           patternMode: Boolean)


  /**
   * Check that `found` conforms to the expected effect in `ctx` and report an error if not.
   * Returns `found` if no error is issued, `bottom` otherwise (to prevent additional spurious errors).
   */
  final def checkConform(found: Effect, tree: Tree, ctx: EffectContext): Effect = ctx.expected match {

    case Some(expected) if !(found <= expected) =>
        //println("reportando error", found, expected )
        ctx.reporter.error(expected, found, tree, ctx.errorInfo)
        bottom
 
    case _ =>
      found
  }


  /**
   * This method implements the effect computation, it returns the effect of `tree`.
   *
   * Concrete effect domains can override this method to assign effects to specific trees.
   */
  def computeEffectImpl(tree: Tree, ctx: EffectContext): Effect = {

    lazy val sym = tree.symbol

    //println("sym:", sym, tree)
    tree match {

      /*** Pattern Matching, some special cases ***/

      case _: CaseDef =>
        //println("->caseDef")
        // @TODO: patternMode is only used for some sanity checks, but still: it should only be enabled
        // for the pattern, maybe for the guard of the CaseDef, but not for the body.
        computeChildEffects(tree, ctx.copy(patternMode = true))

      case Apply(TypeTree(), args) =>
        //println("->caseApply")
        // case class constructor patterns
        // TODO: probably the assertion can fail for `catch` blocks
        assert(ctx.patternMode, s"Apply outside patterns with fun being a TypeTree: $tree")
        computeChildEffects(tree, ctx)

      case UnApply(unapp, args) =>
        //println("->caseUnapply")
        // assert that the unapply is "qual.<unapply-selector>"
        val Apply(_, List(Ident(nme.SELECTOR_DUMMY))) = unapp
        computeApplyEffect(Apply(unapp, args), ctx)


      /*** Method Invocations ***/

      case _: Apply | _: TypeApply =>
        //println("->_: Apply | _: TypeApply")
        computeApplyEffect(tree, ctx)

      case (_: Ident | _: Select) if sym.isMethod =>
        //println("->(_: Ident | _: Select) if sym.isMethod")
        // parameterless methods are applied using a `Select` tree, local ones using an `Ident`
        computeApplyEffect(tree, ctx)


      /*** Modules, Lazy Vals and By-Name Params ***/

      case (_: Ident | _: Select) if (sym.isModule && !sym.isPackage) =>
        //println("->(_: Ident | _: Select) if (sym.isModule && !sym.isPackage)")
        // selection of a module has the effect of the module constructor
        //println("sym:", sym, sym.moduleClass)
        val constr = sym.moduleClass.primaryConstructor
        if (constr == NoSymbol) {
          // modules holding static members of java classes
          assert(sym.isJavaDefined, "could not find primary constructor for module: "+ sym)
          bottom
        } else {
          //val l = neededEffect(tree).getOrElse(latent(constr, Map(), Map(), ctx))
          val latEff = latent(constr, Map(), Map(), ctx)

          val l = updateEffect(tree, latEff, tree.symbol.info.resultType)
          //println("latent;",tree.symbol, latEff, l)
          val treeInfo.Applied(fun, targs, argss) = tree
          l
        }

      case (_: Ident | _: Select) if sym.isLazy =>
        //println("->(_: Ident | _: Select) if sym.isLazy")
        //neededEffect(tree).getOrElse(latent(sym, Map(), Map(), ctx))
        updateEffect(tree, latent(sym, Map(), Map(), ctx), tree.symbol.info.resultType)

      case _: ValDef if sym.isLazy =>
        //println("->_: ValDef if sym.isLazy")
        // lazy val definitions have no effect
        bottom

      case (_: Ident | _: Select) if (sym.isByNameParam || definitions.isByNameParamType(sym.tpe)) =>
        //println("->(_: Ident | _: Select) if (sym.isByNameParam || definitions.isByNameParamType(sym.tpe))")
        // if the enclosing method is annotated `@pure(x)` for by-name parameter `x`, delay the effect
        val isField = sym.isParamAccessor
        val hasRel = ctx.relEnv.exists({
          case RelEffect(ParamLoc(relSym), None) =>
            // for primary constructor expressions, references to by-name parameters are in fact references to the
            // corresponding field. so we need to match the field symbol (sym) to the constr param (relSym)
            sameParam(relSym, sym) || (isField && relSym.name == sym.name && relSym.owner.owner == sym.owner)
          case _ =>
            false
        })
        //aca asume que si no se que funcion es, entonces es top, por ejemplo un //println es Throwable!
        if (hasRel) bottom
        else
            top



      /*** Definitions ***/

      case _: TypeDef | _: DefDef | _: ClassDef | _: ModuleDef | _: Function =>
        //println("->_: TypeDef | _: DefDef | _: ClassDef | _: ModuleDef | _: Function")
        // type, method, class and function definitions have no effect
        bottom


      /*** Otherwise, recurse into Subtrees ***/

      case _ =>
        //println("->_")
        computeChildEffects(tree, ctx)
    }
  }


  /**
   * Computes the effect of `tree` and verifies that it conforms to the expected effect in `ctx` if it
   * is defined. Instead of overriding this methods, effect domains should override `computeEffectImpl`.
   */
  final def computeEffect(tree: Tree, ctx: EffectContext): Effect = {
    //println("COMPUTE EFFECT METHOD", tree, tree.symbol)
    def collectAnnots(typed: Typed): (List[AnnotationInfo], Tree) = {
      val annots = typed.tpt.tpe.annotations
      typed.expr match {
        case typed1: Typed =>
          val (restAnnots, expr) = collectAnnots(typed1)
          (annots ::: restAnnots, expr)
        case expr =>
          (annots, expr)
      }
    }

    if (tree.isErroneous) { println("ERROR"); bottom}
    else {
      //println("matching tree for", tree.symbol)
      val annEffs = annotateEffect(tree)
      
      val e = tree match {
        /* Type ascriptions containing effect annotations AND the `@unchecked` annotations are effect casts,
         * the ascripted tree is not effect-checked.
         *
         * Type ascriptions with effect annotations, but without `@unchecked` are effect ascriptions. They
         * change the expected effect, which allows users to check the effect of a subtree.
         */
        case typed: Typed =>
          //println(tree,"... is typed...")
          val (annots, expr) = collectAnnots(typed)
          //println("la coleccion dio ", annots, "y la expr", expr)
          val hasEffAnnots = existsEffectAnnotation(annots)
          if (hasUncheckedAnnotation(annots) && hasEffAnnots) {
            fromAnnotationList(annots)
          } else if (hasEffAnnots) {
            val newExpected = fromAnnotationList(annots)
            val newExpOpt = Some(newExpected) // or rather ctx.expected.map(_ => newExpected) ?
            //println("computing effect impl")
            val e1 = computeEffectImpl(expr, ctx.copy(expected = newExpOpt))
            checkConform(e1, expr, ctx.copy(expected = newExpOpt))
            newExpected
          } else {
            computeEffectImpl(tree, ctx)
          }

        case _ if annEffs.isDefined =>
          //println("para "+tree, annEffs)
          annEffs.map{ newExpected =>
            val newExpOpt = Some(newExpected)
            val e1 = computeEffectImpl(tree, ctx.copy(expected = newExpOpt))
            checkConform(e1, tree, ctx.copy(expected = newExpOpt))
            newExpected
          }.getOrElse{ computeEffectImpl(tree, ctx) }
          
        case _ =>
          computeEffectImpl(tree, ctx)
      }
      //println("checking conform for ",tree,  tree.symbol, e, ctx)
      //if(tree.symbol!=null)
        //println("el path", buildPath(tree.symbol))

      checkConform(e, tree, ctx)
    }
  }


  /**
   * Computes the effects of all children of `tree`.
   */
  final def computeChildEffects(tree: Tree, ctx: EffectContext) = {
    class ChildEffectsTraverser(ctx: EffectContext) extends Traverser {
      val res = new collection.mutable.ListBuffer[Effect]()

      /* Here we don't call `super.traverse`, i.e. we don't continue into subtrees any deeper */
      override def traverse(tree: Tree) { res += computeEffect(tree, ctx) }

      /* `super.traverse` calls `traverse` for each subtree of `tree` */
      def traverseChilds(tree: Tree): List[Effect] = {
        super.traverse(tree)
        res.toList
      }
    }

    val childEffs = new ChildEffectsTraverser(ctx).traverseChilds(tree)
    joinAll(childEffs: _*)
  }

  def buildPath(sym: Symbol): String = sym.fullName
  def buildPath(tree: Tree): String = {
    val sym = tree.symbol
    //println("enclosing package:",sym.enclosingPackage, sym.enclosingPackage.isEmptyPackage)
    val path = tree match{
      case Select(qualifier, name) if qualifier.symbol != null => 
        //println(qualifier, name, qualifier.symbol, qualifier.symbol.enclosingPackage, buildPackagePath(sym))
        val symb = sym
        buildPackagePath(symb)+symb.enclClass.nameString+"."+name.decoded //println("q:",qualifier, qualifier.symbol.enclosingPackage, sym.enclClass.nameString)
      case Select(New(tpt), name) => tpt.toString+"."+name

      case Ident(name) if sym != null => buildPackagePath(sym)+sym.enclClass.nameString+"."+name.toString
      case DefDef(mods, name, tparams, vparamss, tpt, rhs) if sym != null => buildPackagePath(sym)+sym.enclClass.nameString+"."+name.toString
      case TypeApply(fun, args) =>  buildPackagePath(sym)+sym.enclClass.nameString+"."+fun
      case _ => "<error>"
    }
    path
  }
  def buildPackagePath(sym: Symbol, s: String = ""): String = {
    if(sym != null){
      val pack = sym.enclosingPackage
      if(pack.exists){
        buildPackagePath(pack, pack.nameString+"."+s)
      } else s
    } else s
  }
  def isUncheckedFun_(tree: Tree, l: List[String]) = {
    val path = buildPath(tree)
    //println(tree, tree.symbol, path, l)
    (false /: l)((unchecked, s) => {
      unchecked || s.r.findAllIn(path).hasNext
    })
  }

  def isUncheckedFun(sym: Symbol, l: List[String]) = {
    val path = buildPath(sym)
    //println(tree, tree.symbol, path, l)
    (false /: l)((unchecked, s) => {
      unchecked || s.r.findAllIn(path).hasNext
    })
  }




  object Debugger{
    def identyfySymbol(tree: Tree, indent: Int = 0): Unit = {
      print("  "*indent)
      tree match{
        case Ident(name) =>
          println("Ident(name)")
        case Select(qualifier, selector) =>
          println("Select(qualifier, selector)")
          identyfySymbol(qualifier, indent+1)
        case Apply(fun, args) =>
          println("Apply(fun, args)")
          identyfySymbol(fun, indent+1)
          args.map{identyfySymbol(_, indent+1)}
        case TypeTree() =>
          println("TypeTree()")
        case Literal(value) =>
          println("Literal(value)")
        case This(qual) =>
          println("This(qual)")
        case ValDef(mods, name, tpt, rhs) =>
          println("ValDef(mods, name, tpt, rhs)")
        case DefDef(mods, name, tparams, vparamss, tpt, rhs) =>
          println("DefDef(mods, name, tparams, vparamss, tpt, rhs)")
        case Block(stats, expr) =>
          println("Block(stats, expr)")
        case If(cond, thenp, elsep) =>
          println("If(cond, thenp, elsep)")
        case CaseDef(pat, guard, body) =>
          println("CaseDef(pat, guard, body)")
        case TypeApply(fun, args) =>
          println("TypeApply(fun, args)")

          identyfySymbol(fun, indent+1)
          args.map{identyfySymbol(_, indent+1)}

        case AppliedTypeTree(tpt, args) =>
          println("AppliedTypeTree(tpt, args)")
        case Bind(name, body) =>
          println("Bind(name, body)")
        case Function(vparams, body) =>
          println("Function(vparams, body)")
        case Match(selector, cases) =>
          println("Match(selector, cases)")
        case New(tpt) =>
          println("New(tpt)")
        case Assign(lhs, rhs) =>
          println("Assign(lhs, rhs)")
        case AssignOrNamedArg(lhs, rhs) =>
          println("AssignOrNamedArg(lhs, rhs)")
        case Try(block, catches, finalizer) =>
          println("Try(block, catches, finalizer)")
        case EmptyTree =>
          println("EmptyTree")
        case Throw(expr) =>
          println("Throw(expr)")
        case Super(qual, mix) =>
          println("Super(qual, mix)")
        case TypeBoundsTree(lo, hi) =>
          println("TypeBoundsTree(lo, hi)")
        case Typed(expr, tpt) =>
          println("Typed(expr, tpt)")
        case Import(expr, selectors) =>
          println("Import(expr, selectors)")
        case Template(parents, self, body) =>
          println("Template(parents, self, body)")
        case ClassDef(mods, name, tparams, impl) =>
          println("ClassDef(mods, name, tparams, impl)")
        case ModuleDef(mods, name, impl) =>
          println("ModuleDef(mods, name, impl)")
        case TypeDef(mods, name, tparams, rhs) =>
          println("TypeDef(mods, name, tparams, rhs)")
        case LabelDef(name, params, rhs) =>
          println("LabelDef(name, params, rhs)")
        case PackageDef(pid, stats) =>
          println("PackageDef(pid, stats)")
        case Annotated(annot, arg) =>
          println("Annotated(annot, arg)")
        case SingletonTypeTree(ref) =>
          println("SingletonTypeTree(ref)")
        case SelectFromTypeTree(qualifier, selector) =>
          println("SelectFromTypeTree(qualifier, selector)")
        case CompoundTypeTree(templ) =>
          println("CompoundTypeTree(templ)")
        case ExistentialTypeTree(tpt, whereClauses) =>
          println("ExistentialTypeTree(tpt, whereClauses)")
        case Return(expr) =>
          println("Return(expr)")
        case Alternative(trees) =>
          println("Alternative(trees)")
        case Star(elem) =>
          println("Star(elem)")
        case UnApply(fun, args) =>
          println("UnApply(fun, args)")
        case ArrayValue(elemtpt, trees) =>
          println("ArrayValue(elemtpt, trees)")
        case ApplyDynamic(qual, args) =>
          println("ApplyDynamic(qual, args)")
        case ReferenceToBoxed(idt) =>
          println("ReferenceToBoxed(idt)")
        case _ => println("unknown")
      }
    }
  }
  final class FixedUnderlyingSingleType(pre: Type, sym: Symbol, override val underlying: Type) extends SingleType(pre, sym)
  private def removeAnnotations(tp: Type, cls: List[Symbol]): Type = tp match {
    case AnnotatedType(annots, underlying, _) =>
      underlying.withAnnotations(annots.filterNot(ann => cls.contains(ann.atp.typeSymbol)))

    case st @ SingleType(pre, sym) =>
      val underlying = st.underlying
      val cleanUnderlying = removeAnnotations(underlying, cls)
      if (cleanUnderlying eq underlying) st
      else new FixedUnderlyingSingleType(pre, sym, cleanUnderlying)

    case ExistentialType(quantified, underlying) =>
      val cleanUnderlying = removeAnnotations(underlying, cls)
      if (cleanUnderlying eq underlying) tp
      else ExistentialType(quantified, cleanUnderlying)

    case tp => tp
  }


  /**
   * Apply operation `op` on the result type of (method) type `tp`.
   */
  private def transformResultType(tp: Type, op: Type => Type): Type = tp match {
    case MethodType(args, res) => copyMethodType(tp, args, transformResultType(res, op))
    case PolyType(targs, res) => PolyType(targs, transformResultType(res, op))
    case NullaryMethodType(res) => NullaryMethodType(transformResultType(res, op))
    case tp => op(tp)
  }

  /**
   * Remove all effect annotations from `tp`, or from its result type if it is a method type.
   */
  def removeAllEffectAnnotations(tp: Type) = {
    transformResultType(tp, removeAnnotations(_, allEffectAnnots))
  }

  /**
   * Computes the effect of a method application. This effect consists of three parts:
   *
   *  - the effect of the method selection expression
   *  - the effects of the argument expressions
   *  - the latent (annotated) effect of the method
   *
   * If the function invocation is covered by the relative effect environment in `ctx`, the
   * latent effect does not have to be considered and is therefore `bottom`.
   *
   * Otherwise, all relative effects of the invoked method are expanded using the concrete
   * argument types, which implements effect polymorphism.
   */
  final def computeApplyEffect(tree: Tree, ctx: EffectContext): Effect = {
    val treeInfo.Applied(fun, _, argss) = tree
    val funSym = fun.symbol
    //println("computing apply effect for ", tree)
    // support for while loops: the jump has no effect.
    if (funSym.isLabel){lattice.effectForPureAnnotated}
    else {
      //if(funSym.fullName != buildPath(fun))
      //println("PARA "+tree,buildPath(fun),funSym.moduleClass,funSym.moduleClass.primaryConstructor, funSym.enclClass, funSym.fullName, funSym.name, funSym.enclMethod)
      //println("el simbolo:", funSym, "uncheked:"+ isUncheckedFun(fun, uncheckedFunctions))
      //println(funSym.enclosingPackage.enclosingPackage, funSym.enclosingPackage.exists)
      
 
        //println("no es unchecked")

        val (unaryArgs, repeatedArgs) = splitRepeatedParamss(funSym.paramss, argss)

        // calling computeEffect on `fun` would result in an infinite loop
        //println("aaaaa = ", tree, funSym, fromAnnotation(funSym.info))
        //println("aaaaa = ", tree, fun.symbol, fromAnnotation(fun.symbol.info))
        val funEff   = fun match {
          case Select(qual, _) => 
          //if(!isUncheckedFun(fun.symbol, uncheckedFunctions)){
            computeEffect(qual, ctx)
            //} else{
            //println("FORZANDO EL TREE")
            //bottom
          //}
          case Ident(_) => bottom
        }
        //println("funEff = ", funEff)
        //println("bbbbbb = ", tree, funSym, fromAnnotation(funSym.info))
        //println("bbbbbb = ", tree, fun.symbol, fromAnnotation(fun.symbol.info))

        // for by-name parameters, the effect of the argument expression is only included if the
        // callee has a `@pure(x)` annotation for the by-name parameter x (done by `latent` below)
        val (
          byNameEffs: Map[Symbol, Effect],
          byValEffs: Map[Symbol, Effect],
          repeatedEffs: Map[Symbol, List[Effect]]) = {
          val (byNameParamArgs, byValueParamArgs) = unaryArgs.partition(_._1.isByNameParam)
          // use no expected effect for inferring by-name parameter effects, otherwise effect mismatch
          // errors might be issued unnecessarily
          lazy val noExpectedCtx = ctx.copy(expected = None)
          val byNameEffs0 = byNameParamArgs.map(pa => (pa._1, computeEffect(pa._2, noExpectedCtx)))
          val byValEffs0 = byValueParamArgs.map(pa => (pa._1, computeEffect(pa._2, ctx)))
          val repeatedEffs0 = repeatedArgs.map(pa => (pa._1, pa._2.map(computeEffect(_, ctx))))
          (byNameEffs0, byValEffs0, repeatedEffs0)
        }
        //println("lat acontinuacion")

        val lat = {
          //println("calc lat")
          if (hasRelativeEffect(fun, ctx)) bottom
          else {
            // for non-repeated parameters, we remember the argument type and location in a map and pass it to `latent`.
            // if a parameter has a relative effect, the argument type and location define its expansion.
            val paramLocs = unaryArgs.map(pa => ParamLoc(pa._1))
            //println("paramLocs = ", paramLocs)
            val thisLocMapping = fun match {
              case Select(qual, _) => Some(ThisLoc(funSym.owner) -> argTpeAndLoc(qual))
              case _ => None
            }
            //println("thisLocMapping = ", thisLocMapping)
            //println("latent args = ", funSym, paramLocs.zip(unaryArgs.map(pa => argTpeAndLoc(pa._2))).toMap ++ thisLocMapping, byNameEffs, ctx)
            //println(tree, "parent", tree.symbol.info.finalResultType, removeAllEffectAnnotations(tree.symbol.info.finalResultType) =:= definitions.IntClass.tpe)


            //println(tree)
            val symb: Symbol = tree.symbol


            

            val latEff = latent(funSym, paramLocs.zip(unaryArgs.map(pa => argTpeAndLoc(pa._2))).toMap ++ thisLocMapping, byNameEffs, ctx)
            //println("EPRAPERAPER", tree, symb.info, latEff)
            //neededEffect(tree).getOrElse(latEff)
            val rr = updateEffect(tree, latEff, tree.symbol.info.resultType)
            //println("SETTINGS",tree," TO ",rr)

            rr

          }
        }

        //println("END computing apply effect for ", tree)
        //println("para ", tree, "devuelvo ", funSym, funEff, byValEffs, repeatedEffs, lat, "=",adaptApplyEffect(funSym, funEff, byValEffs, repeatedEffs, lat))
        adaptApplyEffect(funSym, funEff, byValEffs, repeatedEffs, lat)
      
    }

  }

  def adaptApplyEffect(fun: Symbol, funEff: Effect, byValEffs: Map[Symbol, Effect], repeatedEffs: Map[Symbol, List[Effect]], latent: Effect): Effect = {
    ((byValEffs.values ++ repeatedEffs.values.flatten) :\ funEff)(_ u _) u latent
  }


  /**
   * Returns the type and location of a tree which is passed to a method as argument.
   */
  private def argTpeAndLoc(arg: Tree): (Type, Option[Loc]) = {
    val tp = arg.tpe
    arg match {
      case Ident(_) if arg.symbol.owner.isMethod =>
        (tp, Some(ParamLoc(arg.symbol)))
      case This(_) =>
        (tp, Some(ThisLoc(arg.symbol)))
      case _ =>
        (tp, None)
    }
  }

  /**
   * Returns
   *   - a list of non-repeated parameter symbols
   *   - the corresponding list of arguments
   *   - a list of arguments for repeated parameters
   */
  private def splitRepeatedParamss(paramss: List[List[Symbol]], argss: List[List[Tree]]): (Map[Symbol, Tree], Map[Symbol, List[Tree]]) = (paramss, argss) match {
    case (params :: paramss, args :: argss) =>
      val (unaryPairs, repeatedPairs) = params.zipAll(args, NoSymbol, EmptyTree).span(p => !definitions.isRepeated(p._1))

      val (unaryArgs, repeatedArgs) = splitRepeatedParamss(paramss, argss)
      val repeatedArg: Option[(Symbol, List[Tree])] = repeatedPairs match {
        case (sym, arg) :: ps => Some((sym, arg :: ps.map(_._2)))
        case _ => None
      }
      (unaryArgs ++ unaryPairs, repeatedArgs ++ repeatedArg)

//      val (unaryParams, unaryArgs, repeatedArgs) = splitRepeatedParamss(paramss, argss)
//      (unaryPairs.map(_._1) ::: unaryParams, unaryPairs.map(_._2) ::: unaryArgs, repeatedPairs.map(_._2) ::: repeatedArgs)

    case _ =>
      (Map(), Map())
  }

  /**
   * True if the method `fun` exists in the relative effect environment in `ctx`, i.e. if it is
   * a method call on a parameter of an enclosing method, and that method is effect-polymorphic in `fun`.
   */
  private def hasRelativeEffect(fun: Tree, ctx: EffectContext): Boolean = {
    val sym = fun.symbol
    //println(sym, ctx)
    fun match {
      case Select(id @ Ident(_), _) =>
        val r = RelEffect(ParamLoc(id.symbol), Some(sym))
        lteRel(List(r), ctx.relEnv)

      case Select(th @ This(_), _) =>
        val r = RelEffect(ThisLoc(th.symbol), Some(sym))
        lteRel(List(r), ctx.relEnv)

      case _ => false
    }
  }


  /**
   * The latent effect of `fun`, i.e. the maximal effect that might occur when invoking `fun`.
   *
   * @TODO: fixpoint computation needed (rel effects can go in circles!)?
   * @TODO: document what happens here (in general, argtps map specifically)
   *
   * @param argtps A map from the parameters of `fun` to the actual argument types which are passed
   *               in the invocation. These types might be more specific than the parameter types of
   *               `fun`, which implements effect-polymorphism.
   * @param ctx    If `fun` has itself relative effects, but those relative effects are also in the
   *               environment ctx.relEnv (i.e. the enclosing function has the same relative effect), that
   *               relative effect can be ignored.
   */
  private def latent(fun: Symbol, argtps: Map[Loc, (Type, Option[Loc])],
                     byNameEffs: Map[Symbol, Effect], ctx: EffectContext): Effect = {
    //println("calculando latent para "+fun+"="+defaultInvocationEffect(fun))
    defaultInvocationEffect(fun).getOrElse {

      /*if(isUncheckedFun(fun, uncheckedFunctions)){
        Some(effectForPureAnnotated)
      } else*/
      //println("for fun "+ fun)
      val concrete = if(isUncheckedFun(fun, uncheckedFunctions)) bottom else fromAnnotation(fun.info)
      val relEff = relEffects(fun)
      //println("concrete = ", concrete, fun.info, fun.info.finalResultType.annotations)
      //println("relEff = ", relEff)

      val expandedRelEff = relEff map { r =>
        if (lteRelOne(r, ctx.relEnv)) bottom
        else r match {
          case RelEffect(ParamLoc(param), None) if param.isByNameParam =>
            byNameEffs.getOrElse(param, top)

          case RelEffect(paramLoc, funOpt) if (argtps contains paramLoc) =>
            argtps(paramLoc) match {
              case (tp, Some(argLoc)) if lteRelOne(RelEffect(argLoc, funOpt), ctx.relEnv) =>
                // @TODO: document, example. if a parameter is forwarded to another method as parameter, detect rel effects
                bottom
              case (tp, _) =>
                val funSyms = funOpt match {
                  case Some(fun) =>
                    List(tp.member(fun.name).suchThat(m => m.overriddenSymbol(fun.owner) == fun || m == fun))
                  case None =>
                    tp.member(nme.apply).alternatives
                }
                val funEffs = funSyms.map(latent(_, Map(), Map(), ctx))
                joinAll(funEffs: _*)
            }

          case RelEffect(_, Some(fun)) =>
            latent(fun, Map(), Map(), ctx)

          case RelEffect(_, None) =>
            val applyEffs = r.applyMethods.map(latent(_, Map(), Map(), ctx))
            joinAll(applyEffs: _*)
        }
      }
      concrete u (expandedRelEff :\ bottom)(_ u _)
    }
  }

  /**
   * This method is invoked when the effect of a primary constructor is inferred.
   * Effect domains can override it to customize the effect assigned to primary
   * constructors.
   */
  def adaptInferredPrimaryConstrEffect(
      constrDef: DefDef,
      rhsEff: Effect,
      fieldEffs: Map[Symbol, Effect],
      statEffs: List[Effect],
      traitParentConstrEffs: Map[Symbol, Effect]): Effect = {
    (rhsEff /: (fieldEffs.values ++ statEffs ++ traitParentConstrEffs.values))(_ u _)
  }

  def adaptInferredMethodEffect(method: Symbol, eff: Effect) = {
    eff
  }
  
  def adaptExpectedMethodEffect(method: Symbol, eff: Effect) = {
    eff
  }
}

