package effectchecker
package exceptions


import scala.tools.nsc.Global

abstract class ExceptionsDomain extends EffectDomain {
  import global._
  // needs to be a lazy val - otherwise the `global` field of the lattice is set to `null`!
  lazy val lattice = new ExceptionsLattice {
    val global: ExceptionsDomain.this.global.type = ExceptionsDomain.this.global
  }
  import lattice._

  lazy val throwsClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.throws"))
  lazy val annotationClasses: List[Symbol] = List(throwsClass)

  lazy val barTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$bar"))

  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect = {
    def exceptionsOf(tp: Type): List[Type] = tp match {
      case TypeRef(pre, `barTrait`, args) =>
        args.flatMap(exceptionsOf)
      case tp =>
        List(tp)
    }

    //val throwsAnns = annots.filter(_.atp.typeSymbol == throwsClass)
    val anns = annots.filter(annotationClasses contains _.atp.typeSymbol)

    if (anns.isEmpty) {
      default
    } else {

      ((Throws()) /: anns)((eff, annot) => {
        val TypeRef(_, _, List(arg)) = annot.atp
        val tps = {
          // @throws without type arguments is translated to @throws[E] where E is a typeRef to an abstract type symbol
          if (arg.typeSymbol.isAbstractType) List(nothingType)
          else exceptionsOf(arg)
        }
        join(eff, Throws(tps, default.unknown))

      })
    }
  }
  def toAnnotation(eff: Effect): List[AnnotationInfo] = {
    def toType(eff: List[Type]): Type = eff match {
      case Nil => nothingType
      case x :: Nil => x
      case x :: xs => typeRef(barTrait.tpe.prefix, barTrait, List(x, toType(xs)))
    }
    List(AnnotationInfo(typeRef(throwsClass.tpe.prefix, throwsClass, List(toType(eff.tps))), Nil, Nil))
  }


  /**
   * Returns
   *   - the list of types caught by the pattern of `caseDef`
   *   - a boolean indicating if the pattern has been analyzed precisely or not
   */
  def typesMatchingPattern(pattern: Tree): (Throws, Boolean) = pattern match {
    /*** catch one specific exception ***/

    case Bind(_, Typed(Ident(nme.WILDCARD), tpt)) =>
      (Throws(tpt.tpe), true)
    case Typed(Ident(nme.WILDCARD), tpt) =>
      (Throws(tpt.tpe), true)


    /*** catch any exception (i.e. Throwable) ***/

    case Ident(nme.WILDCARD) | Bind(_, Ident(nme.WILDCARD)) =>
      (Throws(throwableType), true)


    /*** alternatives ***/

    case Alternative(trees) =>
      ((Throws(),true) /: trees)((res, tree) => {
        val (mask, isPrecise) = res
        val (treeMask, treeIsPrecise) = typesMatchingPattern(tree)
        (mask u treeMask, isPrecise && treeIsPrecise)
      })

    case _ =>
      (Throws(), false)
  }

  /**
   * Returns
   *   - the list of types which we know are caugth by the case statements `cases`
   *   - a boolean indicating wether the case analysis was precise
   *   - the effect of the righthand-side statements of the cases
   */
  def typesMatchingCases(cases: List[CaseDef]): (lattice.Effect, Boolean, Throws) = {
    ((bottom, true, bottom) /: cases)((res, cdef) => {
      val (mask, isPrecise, catchEff) = res
      val resCatchEff = catchEff
      if (cdef.guard.isEmpty) {
        val (caseMask, caseIsPrecise) = typesMatchingPattern(cdef.pat)
        (mask u caseMask, isPrecise && caseIsPrecise, resCatchEff)
      } else {
        (mask, false, resCatchEff)
      }
    })
  }
  override def adjust(tree: Tree): Tree = {
    val t = super.adjust(tree) //adjust children
    val xi = getEffect(t)

    t match {
      case Try(block, catches, finalizer) =>
        val (mask, maskIsPrecise, catchEff) = typesMatchingCases(catches)

        // set the expected effect: allow masked effects inside the `try`
        val effectsCatched: lattice.Effect = {
          if (maskIsPrecise) xi u mask
          else xi
        }

        //TypeRef(NoPrefix, newTypeName("E1").symbol, Nil)
        //println(TypeTree(newTypeName("E1")))

        //TypeRef(TypeSymbol(final abstract class Unit extends AnyVal))
        /*println(newTermName("X"))
        println(newTypeName("hello.E1"))
        eval
        println(rootMirror.getClassByName(newTypeName("hello.E1")))*/
        /*TODO: Como crear un tipo? */
        //typeTag[]
        //effectsCatched.tps
        adjust(block, effectsCatched, RelEffectList(List())) /*RelEffects is empty */

        //println(mask, maskIsPrecise, catchEff)
        t

      case _ => t
    }
  }

  override def neededEffect(tree: Tree): Option[lattice.Effect] = {
    val eff = tree match {
      case Throw(expr) => Some(Throws(expr.tpe))
      case _ => None
    }
    //println("for ", tree, eff)
    eff
  }

}


abstract class ExceptionsLattice extends EffectLattice {
  val global: Global
  import global._

  lazy val errorType            = rootMirror.getClassByName(newTypeName("java.lang.Error")).tpe
  lazy val runtimeExceptionType = rootMirror.getClassByName(newTypeName("java.lang.RuntimeException")).tpe
  lazy val throwableType        = definitions.ThrowableClass.tpe
  lazy val nothingType          = definitions.NothingClass.tpe

  type Effect = Throws

  type privilege = scala.annotation.effects.Effect

  val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]

  case class Throws(tps: List[Type] = Nil, override val unknown: Boolean = false)  extends WithUnknownEffect{
    def this(tp: Type) = this(List(tp))
    override def toString() = {
      s"@throws[${tps.mkString("|")}]" + (if(unknown) "+¿" else "")
    }
  }
  object Throws {
    def apply(tp: Type) = new Throws(tp)
  }


  lazy val top: Effect    = Throws(throwableType)
  lazy val bottom: Effect = Throws(nothingType)
  lazy val unknown: Effect = new Throws(List(nothingType), true)
  def default = bottom

  def lte(a: Effect, b: Effect): Boolean = lteUnknown(a,b) ||
    a.tps.forall(lteOne(_, b))

  // test for nothingType, that way List(nothingType) conforms to List()
  def lteOne(aTp: Type, b: Effect): Boolean =
    aTp =:= nothingType || b.tps.exists(bTp => aTp <:< bTp)

  def mask(orig: Effect, mask: Effect): Effect =
    Throws(orig.tps.filter(ex =>     // e.g: ex = IOException
      !mask.tps.exists(m =>   //      m  = Exception
        ex <:< m)))           //      since ex <:< m, remove ex

  def elimSub(tps: List[Type]): List[Type] = tps match {
    case Nil => Nil
    case t :: ts =>
      val rest = elimSub(ts.filter(t1 => !(t1 <:< t)))
      if (rest.exists(t1 => t <:< t1)) rest
      else t :: rest
  }
  def join(a: Effect, b: Effect): Effect = {
    Throws(elimSub(a.tps ++ b.tps), joinUnknown(a,b))
  }

  /**
   * think of 'allowed' excepsions
   *   - a allows E1 and E2
   *   - b allows E1 and E3
   *   - the meet(a,b) allows everything which is allowed by both, i.e. E1
   */
  def meet(a: Effect, b: Effect): Effect = {
    val keepA = a.tps.filter(lteOne(_, b))
    val keepB = b.tps.filter(lteOne(_, a))
    // remove duplicates. if a and b both contain some exception type E1,
    // that E1 will end up in both keepA and keepB
    val aNotB = keepA.filterNot(aTp => keepB.exists(bTp => aTp =:= bTp))
    Throws(aNotB ++ keepB, meetUnknown(a,b))
  }

  /* Usable for static effects */
  def diff(a: Effect, b: Effect): Effect = {
    val valid = a.tps.filter(!lteOne(_, b))
    Throws(elimSub(valid))
  }

  /**
   * Removes from `eff` all unchecked exceptions.
   */
  //  def onlyChecked(eff: Effect): Effect =
  //    e.filterNot(ex => ex <:< errorType || ex <:< runtimeExceptionType)



  def getStrictPrivileges(eff: Effect): Effect = {
    /*eff.tps.flatMap{
      tp => Set[privilege](tp)
    }.toSet*/
    Throws(eff.tps)
  }


  def getStrictPrivilegesTree(eff: Effect): List[Tree] = {

    if(eff.tps.size>0){
      val lists = List(
          Apply(Ident(reify{List}.tree.symbol), eff.tps.map{tp => Literal(Constant(tp.toString()))})
        )
      List(Apply(Select(Ident(reify{scala.annotation.effects.runtimeThrows}.tree.symbol), newTermName("apply")),
              List(Apply(Ident(reify{List}.tree.symbol), lists))
              ))
      
    }
    else List[Tree]()

  }


  type OverrideEffect = NeededEffect
  val ctEvidenceOverrideEffect = implicitly[scala.reflect.ClassTag[OverrideEffect]]
  case class NeededEffect(e: Effect)

  def constructNeededEffect(e: Effect): OverrideEffect = NeededEffect(e)
}


