package effectchecker
package ref


import scala.tools.nsc.Global


abstract class RefDomain extends EffectDomain {
  import global._
  // needs to be a lazy val - otherwise the `global` field of the lattice is set to `null`!
  lazy val lattice = new RefLattice {
    val global: RefDomain.this.global.type = RefDomain.this.global
  }
  import lattice._


  lazy val annotationClasses: List[Symbol] = List(readClass,writeClass,allocClass,emptyRefClass/*, unknownClass*/)


  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect = {

    val anns = annots.filter(annotationClasses contains _.atp.typeSymbol)

    if (anns.isEmpty) {
      default
    } else {
      ((RefEffect(emptyRefClass.tpe)) /: anns)((eff, annot) => {
        //val TypeRef(_, _, List(arg)) = annot.atp
        //println("class of ", annot, "is",annot.getClass)
        join(eff, RefEffect(List(annot.atp), default.unknown))
      })
    }
  }

  def toAnnotation(eff: Effect): List[AnnotationInfo] = {
    val l = eff.tps.map{
      tp => AnnotationInfo(tp, Nil, Nil)
    }
    if(eff.unknown)
      l ++ List(AnnotationInfo(unknownClass.tpe, Nil, Nil))
    else l
  }

  override def adjust(tree: Tree): Tree = {
    super.adjust(tree) //adjust children
    val xi = getEffect(tree)

    tree
  }


  override def neededEffect(tree: Tree): Option[lattice.Effect] = {

    tree match {
      case Apply(fun, args) if isAlloc(fun.symbol) => Some(RefEffect(allocClass.tpe))
      case Apply(fun, args) if isWrite(fun.symbol) => Some(RefEffect(writeClass.tpe))
      case Apply(fun, args) if isRead(fun.symbol) => Some(RefEffect(readClass.tpe))
      case _ => None
    }
  }


  /*
  override def delta(tree: Tree): Set[privilege] = {
    val xi = getEffect(tree)
    val needed: Set[privilege] = neededEffect(tree).map{ eff =>
      getStrictPrivileges(eff)
    }.getOrElse(Set[privilege]())
    /* the deltas of the parent plus the one of the domain */
    (super.delta(tree) ++ needed) diff getStrictPrivileges(xi)
  }*/

  private lazy val allocNames = List("alloc").map(newTermName(_))
  private lazy val writeNames = List("write").map(newTermName(_))
  private lazy val readNames = List("read").map(newTermName(_))


  private def isAlloc(sym: Symbol) = {
    // observed NoSymbol in some situations (annotation constructor invocations)
    //println("symbolo se llama:", sym.name,sym != NoSymbol,sym.owner == PredefMClass)
    sym != NoSymbol && allocNames.contains(sym.name)
  }
  private def isWrite(sym: Symbol) = {
    // observed NoSymbol in some situations (annotation constructor invocations)
    sym != NoSymbol && writeNames.contains(sym.name)
  }
  private def isRead(sym: Symbol) = {
    // observed NoSymbol in some situations (annotation constructor invocations)
    sym != NoSymbol && readNames.contains(sym.name)
  }

}

abstract class RefLattice extends EffectLattice {
  val global: Global
  import global._

  lazy val readClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.read"))
  lazy val writeClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.write"))
  lazy val allocClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.alloc"))
  lazy val emptyRefClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.emptyRef"))


  type Effect = RefEffect
  //type privilege = scala.annotation.effects.Effect
  val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]

  case class RefEffect(tps: List[Type] = Nil, override val unknown: Boolean = false) extends WithUnknownEffect {
    def this(tp: Type) = this(List(tp))
    override def toString() = {
      tps.map{tp => "@"+tp.toString()}.mkString(" ") + (if(unknown) "+¿" else "")
    }


  }
  object RefEffect {
    def apply(tp: Type) = new RefEffect(tp)
  }


  lazy val top: Effect    = RefEffect(List(readClass.tpe, writeClass.tpe, allocClass.tpe))
  lazy val bottom: Effect = RefEffect(List(emptyRefClass.tpe))
  lazy val unknown: Effect = new RefEffect(List(emptyRefClass.tpe), true)
  def default = bottom

  def lte(a: Effect, b: Effect): Boolean = lteUnknown(a,b) ||
      /*a == bottom ||*/
      a.tps.forall(lteOne(_, b))

  // test for emptyRef, that way List(emptyRef) conforms to List()
  def lteOne(aTp: Type, b: Effect): Boolean =
    aTp =:= emptyRefClass.tpe || b.tps.exists(bTp => aTp <:< bTp)

  def elimSub(tps: List[Type]): List[Type] = tps match {
    case Nil => Nil
    case t :: ts =>
      val rest = elimSub(ts.filter(t1 => !(t1 <:< t)))
      if (rest.exists(t1 => t <:< t1)) rest
      else t :: rest
  }

  def join(a: Effect, b: Effect): Effect = {
    val tps = if(a == bottom) b.tps
    else if(b == bottom) a.tps
    else a.tps ++ b.tps
    RefEffect(elimSub(tps), joinUnknown(a,b))

  }


  /**
   * think of 'allowed' exceptions
   *   - a allows E1 and E2
   *   - b allows E1 and E3
   *   - the meet(a,b) allows everything which is allowed by both, i.e. E1
   */
  def meet(a: Effect, b: Effect): Effect = {
    val keepA = a.tps.filter(lteOne(_, b))
    val keepB = b.tps.filter(lteOne(_, a))
    // remove duplicates. if a and b both contain some exception type E1,
    // that E1 will end up in both keepA and keepB
    val aNotB = keepA.filterNot(aTp => keepB.exists(bTp => aTp =:= bTp))
    RefEffect(aNotB ++ keepB, meetUnknown(a,b))
  }
  /* Usable for static effects */
  def diff(a: Effect, b: Effect): Effect = {
    RefEffect(elimSub(a.tps diff b.tps))
  }

  def getStrictPrivileges(eff: Effect): Effect = {
    RefEffect(eff.tps)
  }

  def getStrictPrivilegesTree(eff: Effect): List[scala.tools.nsc.Global#Tree] = {
    eff.tps.flatMap{ tp =>
      if(tp == readClass.tpe) List(reify{scala.annotation.effects.read()}.tree)
      else if(tp == writeClass.tpe) List(reify{scala.annotation.effects.write()}.tree)
      else if(tp == allocClass.tpe) List(reify{scala.annotation.effects.alloc()}.tree)
      else List()
    }.toList
  }

  type OverrideEffect = NeededEffect
  val ctEvidenceOverrideEffect = implicitly[scala.reflect.ClassTag[OverrideEffect]]
  case class NeededEffect(e: Effect)

  def constructNeededEffect(e: Effect): OverrideEffect = NeededEffect(e)
}
